UNIT Algebra;

{$I OsWahl.inc}

{$IFDEF DOS} {$M 65520,0,655360} {$N+}
{$ELSE} {$mode objfpc} {$H+}
        {$IFDEF X86_64} {$MAXSTACKSIZE $7FFFFFF} {$ENDIF}
{$ENDIF}

INTERFACE

USES Stdio;

CONST VERSION_INFO_Algebra = 'Algebra   V 2.43 '
{$IFDEF LINUX_32} + '(Linux 32)' {$ENDIF} {$IFDEF LINUX_64} + '(Linux 64)' {$ENDIF}
{$IFDEF WIN_32}   + '(32bit)'    {$ENDIF} {$IFDEF WIN_64}   + '(64bit)'    {$ENDIF}
{$IFDEF BIT_16}   + '(16bit)'    {$ENDIF} + ' 02.04.2025   (c) ALFWARE B.Schubert';

(*************************************************************************)
(* Typ ZAHL und Definitionen                                             *)
(*************************************************************************)

CONST  MAXZIFFER = {$IFDEF DOS} 3000 {$ELSE} 3000 {$ENDIF};
(* 32/64bit bis 1.000.000 m_oeglich aber dann viel langsamer! *)
TYPE   ZAHL = RECORD
         plus:BOOLEAN; anz,potenz:INTEGER;
         ziff:ARRAY[1..MAXZIFFER] OF BYTE;
       END;

VAR alEins,alNull,alZwei,alMinusEins:ZAHL;

PROCEDURE alWrite(z:ZAHL);
(* schreibt den Typ Zahl, mit 3er-Punkt und E *)

PROCEDURE alSpare(VAR z:ZAHL; paket:BOOLEAN);
(* potenz=TRUE:  packe z ein
   potenz=FALSE: packe z aus *)

PROCEDURE alStr2Zahl(s:STRING; VAR z:ZAHL);
(* wandelt Zeichenkette in Zahl *)

PROCEDURE alZahl2Str(z:ZAHL; VAR s:STRING);
(* wandelt Zahl in Zeichenkette *)

PROCEDURE alCard2Zahl(c:CARD; VAR z:ZAHL);
(* wandelt Card in Zahl *)

PROCEDURE alZahl2Card(z:ZAHL; VAR c:CARD);
(* wandelt Zahl in Card *)

FUNCTION alGT(a,b:ZAHL):BOOLEAN;
(* a > b *)

FUNCTION alGTabs(a,b:ZAHL):BOOLEAN;
(* |a| > |b| *)

FUNCTION alEQ(a,b:ZAHL):BOOLEAN;
(* a = b *)

FUNCTION alHat0(z:ZAHL):INTEGER;
(* z hat ... Nullen *)

FUNCTION alIs0(z:ZAHL):BOOLEAN;
(* z ist Null *)

FUNCTION alIs1(z:ZAHL):BOOLEAN;
(* z ist Eins *)

PROCEDURE alNeg(VAR z:ZAHL);
(* z = -z *)

PROCEDURE alInc(VAR z:ZAHL);
(* z = z + 1 *)

PROCEDURE alDec(VAR z:ZAHL);
(* z = z - 1 *)

PROCEDURE alAdd(a,b:ZAHL; VAR e:ZAHL);
(* e = a + b *)

PROCEDURE alSub(a,b:ZAHL; VAR e:ZAHL);
(* e = a - b *)

PROCEDURE alMul(a,b:ZAHL; VAR e:ZAHL);
(* e = a * b *)

PROCEDURE alDiv(a,b:ZAHL; VAR e:ZAHL);
(* e = a DIV b *)

PROCEDURE alDivR(a,b:ZAHL; VAR e,r:ZAHL);
(* e = a DIV b, r = a MOD b *)

PROCEDURE alMod(a,b:ZAHL; VAR e:ZAHL);
(* e = a MOD b *)

PROCEDURE alMod2(a,b:ZAHL; VAR e:ZAHL);
(* langsameres MOD: e = a - b * (a DIV b) *)

PROCEDURE alMaln(a:ZAHL; n:INTEGER; VAR e:ZAHL);
(* e = n * z *)

PROCEDURE alPot(a,b:ZAHL; VAR e:ZAHL);
(* e = a ^ b *)

PROCEDURE alFak(n:ZAHL; VAR e:ZAHL);
(* e = n! *)

PROCEDURE alPot10(a:ZAHL; n:INTEGER; VAR e:ZAHL);
(* e = a * 10^b *)

PROCEDURE alSHL(a:ZAHL; n:INTEGER; VAR e:ZAHL);
(* e = a << n   Linksschieben *)

PROCEDURE alSHR(a:ZAHL; n:INTEGER; VAR e:ZAHL);
(* e = a >> n   Rechtsschieben *)

(*************************************************************************)
(* Typ BITSET (Mengen) und Definitionen                                  *)
(*************************************************************************)

TYPE  BITSET = SET OF BYTE;

CONST  bNull = 0; bMin = 1;  bMax = 255;

VAR bLeer,bVoll,b0255: BITSET; (* bVoll=[1..255] bLeer=[] b0255=[0..255] *)

(* Achtung! Obwohl der Typ Bitset auch fuer das "0"-Element definiert ist,
   wird dieses bei den Funktionen BCARD, (O)BNUM und (O)BCOMP nicht
   mitgezaehlt!
   OBCOMP und OBNUM sind schnellere Alternativen zu BCOMP und BNUM -
   allerdings sollte man auf Spielchen mit [] usw. verzichten, denn
   es werden nicht alle Fehlermoeglichkeiten abgefangen                  *)

FUNCTION bIst(m:BITSET; nr:BYTE):BOOLEAN;
(* ist nr in m enthalten ?  nur "IN" ist aber 4x schneller! *)

FUNCTION bCard(m:BITSET):BYTE;
(* wieviele Elemente hat m ? *)

FUNCTION bComp(m:BITSET; nr:BYTE):BYTE;
(* wie heisst die nr. Komponente von m *)

FUNCTION bNum(m:BITSET; nr:BYTE):BYTE;
(* die wievielte Komponente von m ist nr *)

FUNCTION ObComp(m:BITSET; nr:BYTE):BYTE;
(* das schnelle bComp, prueft nicht auf Sinnvollkeit *)

FUNCTION ObNum(m:BITSET; nr:BYTE):BYTE;
(* das schnelle bNum, prueft nicht auf Sinnvollkeit *)

PROCEDURE bVerein(VAR e:BITSET; m,n:BITSET);
(* Vereinungsmenge *)

PROCEDURE bDurch(VAR e:BITSET; m,n:BITSET);
(* Durchschnittsmenge *)

PROCEDURE bIncl(VAR m:BITSET; nr:BYTE);
(* m+[nr] *)

PROCEDURE bExcl(VAR m:BITSET; nr:BYTE);
(* m-[nr] *)

PROCEDURE bInit(VAR m:BITSET; von, bis:BYTE);
(* m enthlt [von, bis] *)

PROCEDURE b0Init(VAR m:BITSET; nr:BYTE);
(* m enthlt [0 .. bis] *)

PROCEDURE b1Init(VAR m:BITSET; nr:BYTE);
(* m enthlt [1 .. bis] *)

PROCEDURE bWrite(m:BITSET; sep:CHAR);
(* Ausgabe m (Write) *)

PROCEDURE bWriteF(VAR d:TEXT; m:BITSET; sep:CHAR);
(* Ausgabe m ( Write(datei) ) *)

IMPLEMENTATION

(*************************************************************************)
(* Funktionen Implementation f_uer den Typ ZAHL                          *)
(*************************************************************************)

PROCEDURE alWrite(z:ZAHL);
VAR i:INTEGER;
BEGIN
  IF NOT z.plus AND ((z.anz>1) OR (z.ziff[1]<>0)) THEN Write('-');
  FOR i:=z.anz DOWNTO 2 DO IF i<=z.anz THEN BEGIN
    Write(z.ziff[i]); IF i mod 3 = 1 THEN Write('.');
  END;
  Write(z.ziff[1]); IF z.potenz<>0 THEN Write(' E ',z.potenz);
END (* alWrite *);

PROCEDURE alSpare(VAR z:ZAHL; paket:BOOLEAN);
BEGIN IF paket THEN alPot10(z,alHat0(z),z) ELSE alPot10(z,-z.potenz,z);
END (* alSpare *);

PROCEDURE alStr2Zahl(s:STRING; VAR z:ZAHL);
VAR i:INTEGER;
BEGIN
  z.plus:=NOT (s[1]='-'); z.potenz:=0;
  WHILE (s<>'') AND NOT (s[1] in ['1'..'9']) DO s:=COPY(s,2,LENGTH(s)-1);
  IF s='' THEN s:='0'; z.anz:=LENGTH(s);
  FOR i:=1 to z.anz DO z.ziff[i]:=ORD(s[z.anz-i+1])-48;
END (* alStr2Zahl *);

PROCEDURE alZahl2Str(z:ZAHL; VAR s:STRING);
VAR i:INTEGER;
BEGIN
  s:=CHR(48+z.ziff[1]);
  FOR i:=2 to z.anz DO s:=CHR(48+z.ziff[i])+s;
  IF NOT z.plus THEN s:='-'+s;
END (* alZahl2Str *);

PROCEDURE alCard2Zahl(c:CARD; VAR z:ZAHL);
VAR s:STRING;
BEGIN
  STR(c,s);
  alStr2Zahl(s,z);
END (* alCard2Zahl *);

PROCEDURE alZahl2Card(z:ZAHL; VAR c:CARD);
VAR i:INTEGER; 
BEGIN
  c:=z.ziff[z.anz];
  FOR i:=z.anz-1 downto 1 DO c:=10*c+z.ziff[i];
  IF NOT z.plus THEN c:=-c;
END (* alZahl2Card *);

FUNCTION alGT(a,b:ZAHL):BOOLEAN;
VAR gt,gt0:BOOLEAN; i:INTEGER;
BEGIN
  IF a.plus<>b.plus THEN BEGIN alGT:=(a.plus>b.plus); EXIT; END;
  IF a.anz<>b.anz THEN BEGIN
    IF a.plus THEN alGT:=(a.anz>b.anz) ELSE alGT:=(a.anz<b.anz); EXIT;
  END;
  i:=a.anz; gt:=FALSE; gt0:=gt; alGT:=gt;
  REPEAT
    IF a.ziff[i]<>b.ziff[i] THEN BEGIN
      gt:=(a.ziff[i]>b.ziff[i]); gt0:=TRUE; alGT:=(gt=a.plus);
    END;
    DEC(i);
  UNTIL (i=0) OR gt0;
END (* alGT *);

FUNCTION alGTabs(a,b:ZAHL):BOOLEAN;
VAR z:ZAHL;
BEGIN
  z:=b; z.plus:=a.plus;
  alGTabs:=alGT(a,z);
END (* alGTabs *);

FUNCTION alEQ(a,b:ZAHL):BOOLEAN;
VAR eq:BOOLEAN; i:INTEGER;
BEGIN
  IF (a.anz<>b.anz) OR (a.plus<>b.plus) THEN BEGIN alEQ:=FALSE; EXIT END;
  i:=a.anz; eq:=TRUE;
  REPEAT
    IF a.ziff[i]<>b.ziff[i] THEN eq:=FALSE; DEC(i);
  UNTIL (i=0) OR NOT eq;
  alEQ:=eq;
END (* alEQ *);

FUNCTION alHat0(z:ZAHL):INTEGER;
VAR j:INTEGER;
BEGIN
  j:=1; WHILE (j<=z.anz) AND (z.ziff[j]=0) DO INC(j);
  alHat0:=j-1;
END (* alHat0 *);

FUNCTION alIs0(z:ZAHL):BOOLEAN;
BEGIN alIs0:=(z.anz=1) AND (z.ziff[1]=0);
END (* alIs0 *);

FUNCTION alIs1(z:ZAHL):BOOLEAN;
BEGIN alIs1:=(z.anz=1) AND (z.ziff[1]=1);
END (* alIs1 *);

PROCEDURE alNeg(VAR z:ZAHL);
BEGIN z.plus:=NOT z.plus;
END (* alNeg *);

PROCEDURE alInc(VAR z:ZAHL);
BEGIN alAdd(z,alEins,z);
END (* alInc *);

PROCEDURE alDec(VAR z:ZAHL);
BEGIN alSub(z,alEins,z);
END (* alDec *);

PROCEDURE alAdd0(a,b:ZAHL; VAR e:ZAHL);
VAR i:INTEGER; s,ue:BYTE;
BEGIN
  ue:=0;
  FOR i:=a.anz DOWNTO b.anz+1 DO b.ziff[i]:=0;
  FOR i:=1 to a.anz DO BEGIN
    s:=ue+a.ziff[i]+b.ziff[i];
    IF s>9 THEN BEGIN s:=s-10; ue:=1; END ELSE ue:=0;
    e.ziff[i]:=s;
  END;
  e.ziff[a.anz+1]:=ue; e.anz:=a.anz+ue; e.plus:=a.plus; e.potenz:=0;
END (* alAdd0 *);

PROCEDURE alSub0(a,b:ZAHL; VAR e:ZAHL);
VAR i:INTEGER;
BEGIN
  FOR i:=a.anz DOWNTO b.anz+1 DO b.ziff[i]:=0;
  FOR i:=1 to a.anz-1 DO BEGIN
    IF b.ziff[i]>a.ziff[i]
      THEN BEGIN e.ziff[i]:=10+a.ziff[i]-b.ziff[i]; INC(b.ziff[i+1]); END
      ELSE e.ziff[i]:=a.ziff[i]-b.ziff[i];
  END;
  e.ziff[a.anz]:=a.ziff[a.anz]-b.ziff[a.anz];
  e.anz:=a.anz; e.plus:=a.plus; e.potenz:=0;
  WHILE (e.anz>1) AND (e.ziff[e.anz]=0) DO DEC(e.anz);
END (* alSub0 *);

PROCEDURE alAdd(a,b:ZAHL; VAR e:ZAHL);
BEGIN
  IF alGTabs(b,a)
    THEN
      IF a.plus=b.plus
        THEN BEGIN alAdd0(b,a,e); e.plus:=a.plus; END
        ELSE BEGIN alSub0(b,a,e); e.plus:=b.plus; END
    ELSE BEGIN
      IF a.plus=b.plus
        THEN alAdd0(a,b,e)
        ELSE alSub0(a,b,e);
      e.plus:=a.plus
END END (* alAdd *);

PROCEDURE alSub(a,b:ZAHL; VAR e:ZAHL);
BEGIN
  IF alGTABs(b,a)
    THEN
      IF a.plus=b.plus
        THEN BEGIN alSub0(b,a,e); e.plus:=NOT a.plus; END
        ELSE BEGIN alAdd0(b,a,e); e.plus:=a.plus; END
    ELSE BEGIN
      IF a.plus=b.plus
        THEN alSub0(a,b,e)
        ELSE alAdd0(a,b,e);
      e.plus:=a.plus
END END (* alSub *);

PROCEDURE alMul(a,b:ZAHL; VAR e:ZAHL);
VAR i:INTEGER; x:ZAHL;
BEGIN
  e:=alNull;
  IF b.anz>a.anz THEN BEGIN x:=a; a:=b; b:=x; END;
  FOR i:=1 to b.anz DO BEGIN
    alMaln(a,b.ziff[i],x);
    alSHL(x,i-1,x);
    alADD(e,x,e);
  END;
  e.potenz:=a.potenz+b.potenz; e.plus:=(a.plus=b.plus);
END (* alMul *);

PROCEDURE alDiv(a,b:ZAHL; VAR e:ZAHL);
VAR j,i:INTEGER;
BEGIN
  IF alIs0(a) AND alIs0(b) THEN e:=alEins ELSE e:=alNull;
  IF alGTABS(b,a) OR alIs0(b) THEN EXIT;
  j:=a.anz-b.anz; e.potenz:=0; e.plus:=(a.plus=b.plus);
  alSHL(b,j,b); a.plus:=TRUE; b.plus:=TRUE;
  FOR i:=j DOWNTO 0 DO BEGIN
    WHILE NOT alGTABS(b,a) DO BEGIN INC(e.ziff[1]); alSUB(a,b,a); END;
    IF i>0 THEN BEGIN alSHR(b,1,b); alSHL(e,1,e); END;
  END;
  (* es k_oennten Vornullen entstanden sein *)
  WHILE (e.anz>1) AND (e.ziff[e.anz]=0) DO DEC(e.anz);
END (* alDiv *);

PROCEDURE alDivR(a,b:ZAHL; VAR e,r:ZAHL);
VAR j,i:INTEGER;
BEGIN
  IF alIs0(a) AND alIs0(b) THEN e:=alEins ELSE e:=alNull;
  IF alGTABS(b,a) OR alIs0(b) THEN EXIT;
  j:=a.anz-b.anz; e.potenz:=0; e.plus:=(a.plus=b.plus);
  alSHL(b,j,b); a.plus:=TRUE; b.plus:=TRUE;
  FOR i:=j DOWNTO 0 DO BEGIN
    WHILE NOT alGTABS(b,a) DO BEGIN INC(e.ziff[1]); alSUB(a,b,a); END;
    IF i>0 THEN BEGIN alSHR(b,1,b); alSHL(e,1,e); END;
  END;
  r:=a;
  (* es k_oennten Vornullen entstanden sein *)
  WHILE (e.anz>1) AND (e.ziff[e.anz]=0) DO DEC(e.anz);
  WHILE (r.anz>1) AND (r.ziff[r.anz]=0) DO DEC(r.anz);
END (* alDivR *);

PROCEDURE alMod(a,b:ZAHL; VAR e:ZAHL);
VAR j,i:INTEGER;
BEGIN
  e:=alNull;
  IF alIs0(a) OR alIs0(b) OR alGTABS(b,a) OR (NOT a.plus) OR (NOT b.plus)
    THEN BEGIN e:=a; EXIT; HALT; END;
  j:=a.anz-b.anz; alSHL(b,j,b); a.plus:=TRUE; b.plus:=TRUE;
  FOR i:=j DOWNTO 0 DO BEGIN
    WHILE NOT alGTABS(b,a) DO BEGIN INC(e.ziff[1]); alSUB(a,b,a); END;
    IF i>0 THEN BEGIN alSHR(b,1,b); alSHL(e,1,e); END;
  END;
  e:=a; e.plus:=TRUE; e.potenz:=0;
  (* es k_oennten Vornullen entstanden sein *)
  WHILE (e.anz>1) AND (e.ziff[e.anz]=0) DO DEC(e.anz);
END (* alMod *);

PROCEDURE alMod2(a,b:ZAHL; VAR e:ZAHL);
VAR q,p:ZAHL;
BEGIN
  alDiv(a,b,q);
  alMul(b,q,p);
  alSub(a,p,e);
END (* alMod2 *);

PROCEDURE alMaln(a:ZAHL; n:INTEGER; VAR e:ZAHL);
VAR i:INTEGER;
BEGIN
  IF n<=0 THEN e:=alNull ELSE e:=a;
  FOR i:=1 to n-1 DO alAdd(e,a,e);
END (* alMaln *);

PROCEDURE alPot10(a:ZAHL; n:INTEGER; VAR e:ZAHL);
BEGIN
  IF (n=0) OR ((n=1) AND (a.anz=1)) THEN BEGIN e:=a; exit; END;
  IF n>0 THEN BEGIN alSHR(a,n,e); e.potenz:=a.potenz+n END;
  IF n<0 THEN BEGIN alSHL(a,-n,e); e.potenz:=a.potenz+n END;
END (* alPot10 *);

PROCEDURE alPot(a,b:ZAHL; VAR e:ZAHL);
BEGIN
  e:=alEins;
  WHILE(alGT(b,alNull)) DO BEGIN
    alMUL(e,a,e); alDec(b);
  END;
END (* alPot *);

PROCEDURE alFak(n:ZAHL; VAR e:ZAHL);
BEGIN
  e:=alEins;
  WHILE(alGT(n,alEins)) DO BEGIN
    alMUL(e,n,e); alDec(n);
  END;
END (* alFak *);

PROCEDURE alSHL(a:ZAHL; n:INTEGER; VAR e:ZAHL);
VAR i:INTEGER;
BEGIN
  IF n<=0 THEN BEGIN e:=a; exit; END;
  FOR i:=a.anz DOWNTO 1 DO BEGIN e.ziff[i+n]:=a.ziff[i]; e.ziff[i]:=0; END;
  e.anz:=a.anz+n; e.plus:=a.plus; e.potenz:=a.potenz;
END (* alSHL *);

PROCEDURE alSHR(a:ZAHL; n:INTEGER; VAR e:ZAHL);
VAR i:INTEGER;
BEGIN
  IF n<=0 THEN BEGIN e:=a; exit; END;
  FOR i:=1 to a.anz DO e.ziff[i]:=a.ziff[i+n];
  FOR i:=0 to n-1 DO e.ziff[a.anz-i]:=0;
  IF a.anz<=n THEN e.anz:=1 ELSE e.anz:=a.anz-n;
  e.plus:=a.plus; e.potenz:=a.potenz;
END (* alSHR *);

(*************************************************************************)
(* Funktionen Implementation f_uer den Typ BITSET                        *)
(*************************************************************************)

FUNCTION bIst(m:BITSET; nr:BYTE):BOOLEAN;
BEGIN bIst:=(nr IN m);
END (* bIst *);

FUNCTION bCard(m:BITSET):BYTE;
VAR b,s:BYTE;
BEGIN
  IF m=[] THEN bCard:=0 ELSE IF m=bvoll THEN bCard:=255
  ELSE BEGIN
    s:=0; FOR b:=bMin TO bMax DO IF b IN m THEN INC(s);
    bCard:=s;
END END (* bCard *);

FUNCTION bComp(m:BITSET; nr:BYTE):BYTE;
VAR b,s:BYTE;
BEGIN
  s:=bCard(m); IF (nr>s)  OR (nr=0) THEN bComp:=0
  ELSE BEGIN
    s:=nr; b:=bMin;
    REPEAT IF b in m THEN DEC(s); INC(b) UNTIL s=0;
    bComp:=b-1
END END (* bComp *);

FUNCTION bNum(m:BITSET; nr:BYTE):BYTE;
VAR b,s:BYTE;
BEGIN
  s:=bCard(m);
  IF (s*nr=0) OR (NOT (nr IN m)) THEN bNum:=0
  ELSE BEGIN
    s:=bMin; FOR b:=bMin TO nr-1 DO IF b IN m THEN INC(s);
    bNum:=s
END END (* bNum *);

FUNCTION ObComp(m:BITSET; nr:BYTE):BYTE;
VAR b,s:BYTE;
BEGIN
  s:=nr; b:=bMin;
  REPEAT IF b in m THEN DEC(s); INC(b) UNTIL s=0;
  ObComp:=b-1
END (* ObComp *);

FUNCTION ObNum(m:BITSET; nr:BYTE):BYTE;
VAR b,s:BYTE;
BEGIN
  s:=bMin; FOR b:=bMin TO nr-1 DO IF b IN m THEN INC(s);
  ObNum:=s
END (* ObNum *);

PROCEDURE bVerein(VAR e:BITSET; m,n:BITSET);
BEGIN e:=m+n
END (* bVerein *);

PROCEDURE bDurch(VAR e:BITSET; m,n:BITSET);
BEGIN e:=m*n
END (* bDurch *);

PROCEDURE bIncl(VAR m:BITSET; nr:BYTE);
BEGIN m:=m+[nr];
END (* bIncl *);

PROCEDURE bExcl(VAR m:BITSET; nr:BYTE);
BEGIN m:=m-[nr];
END (* bExcl *);

PROCEDURE bInit(VAR m:BITSET; von, bis:BYTE);
VAR bi: BYTE;
BEGIN m:=[]; FOR bi:=von TO bis DO m:=m+[bi];
END (* bInit *);

PROCEDURE b0Init(VAR m:BITSET; nr:BYTE);
BEGIN m:=[0..nr];
END (* b0Init *);

PROCEDURE b1Init(VAR m:BITSET; nr:BYTE);
BEGIN m:=[1..nr];
END (* b1Init *);

PROCEDURE bWrite(m:BITSET; sep:CHAR);
VAR b:BYTE; k:BOOLEAN;
BEGIN
  IF sep<>' ' THEN Write('['); k:=TRUE;
  FOR b:=bNull TO bMax DO IF b IN m THEN
    IF k THEN BEGIN Write(b); k:=FALSE END ELSE Write(sep,b);
  IF sep<>' ' THEN Write(']');
END (* bWrite *);

PROCEDURE bWriteF(VAR d:TEXT; m:BITSET; sep:CHAR);
VAR b:BYTE; k:BOOLEAN;
BEGIN
  IF sep<>' ' THEN Write(d,'['); k:=TRUE;
  FOR b:=bNull TO bMax DO IF b IN m THEN
    IF k THEN BEGIN Write(d,b); k:=FALSE END ELSE Write(d,sep,b);
  IF sep<>' ' THEN Write(d,']');
END (* bWriteF *);

BEGIN
  alStr2Zahl('1',alEins); alStr2Zahl('0',alNull); alStr2Zahl('2',alZwei);
  alCard2Zahl(-1,alMinusEins);
  bLeer:=[]; b0Init(b0255, 255); bVoll:=[1..255];
END (* Algebra.TPU *).
