UNIT Zeit;

{$I OsWahl.inc}

{$IFDEF DOS} {$M 65520,0,655360} {$N+}
{$ELSE} {$mode objfpc} {$H+}
        {$IFDEF WIN_64} {$MAXSTACKSIZE $7FFFFFF} {$ENDIF}
        {$DEFINE ZEITOPT}
{$ENDIF}

INTERFACE

USES
  Stdio, DOS, Watch
  {$IFDEF BIT_64}, SysUtils, DateUtils {$ENDIF}
  {$IFDEF LINUX_64}, CWString, CLocale {$ENDIF};

CONST
  VERSION_INFO_Zeit = 'Zeit   V 2.43 '
  {$IFDEF LINUX_32} + '(Linux 32)' {$ENDIF} {$IFDEF LINUX_64} + '(Linux 64)' {$ENDIF}
  {$IFDEF WIN_32}   + '(32bit)'    {$ENDIF} {$IFDEF WIN_64}   + '(64bit)'    {$ENDIF}
  {$IFDEF BIT_16}   + '(16bit)'    {$ENDIF} + ' 31.03.2025   (c) ALFWARE B.Schubert';

TYPE
  Moment = RECORD
    wjahr, wmonat, wtag,
    wstunde, wminute, wsekunde,
    wmsek, wwotag                : WORD;
    jahr, monat, tag,
    stunde, minute, sekunde,
    msek, wotag                  : LONGINT;
    abstag                       : LONGINT;
    abssek                       : DOUBLE;
    {$IFNDEF BIT_64}
    absMS                        : DOUBLE;
    {$ELSE}
    absDT                        : TDateTime;
    absTS                        : TTimeStamp;
    absMS                        : CARD;
    {$ENDIF}
    TSDos                        : TDosTime;
    TSUnx                        : TUnxTime;
    TSWin                        : TWinTime;
    heute_feiertag               : INTEGER;
    welcher_feiertag             : ARRAY [1..5] OF STRING[30];
  END;

  DiffMoment = RECORD
    ijahr, imonat, itag,
    istunde, iminute, isekunde,
    imsek                        : LONGINT;
    rjahr, rmonat, rwoche,
    rtag, rstunde, rminute,
    rsekunde                     : DOUBLE;
  END;

VAR
  Akt_Moment, Test_Moment        : Moment;
  Diff_Moment                    : DiffMoment;
  {$IFNDEF BIT_64}
  Summe_Moment                   : DiffMoment;
  {$ENDIF}
  Formate: RECORD
    TT_MM_JJ, TT_MM_JJJJ,
    JJJJ_MM_TT, HH_MM_SS,
    HH_MM_SS_MS, Wochentag,
    Abs_MS, Abs_Tag, Datum_Zeit,
    Timestamp, FileStampDT,
    Ddatum, DdatumTzeit,
    {$IFDEF BIT_64}
    Abs_DT, Abs_TS,
    {$ENDIF}
    TS_Win, TS_Dos, TS_Unx       : STRING;
  END;

  Diff_TimeStamp, Diff_TimeStamp24,
  {$IFNDEF BIT_64}
  Summe_TimeStamp, Summe_TimeStamp24,
  {$ENDIF}
  Diff_Einzel, Diff_ABS,
  Diff_Abs_Jahre, Diff_Abs_Monate,
  Diff_Abs_Wochen, Diff_Abs_Tage,
  Diff_Abs_Stunden, Diff_Abs_Minuten,
  Diff_Abs_Sekunden              : STRING;

FUNCTION StopUhr(Kommentar: STRING): STRING;
(* Neue StechUhr TimeStamp als Ausgabestring *)

FUNCTION Intervall_Start(nr: INTEGER): STRING;
(* Programmstart-Zeitpunkt *)

FUNCTION Intervall_Stop(nr: INTEGER): STRING;
(* Programmende-Zeitpunkt *)

FUNCTION Intervall_Dauer(nr: INTEGER): STRING;
(* Programmdauer Stop minus Start *)

FUNCTION Intervall_Dauer24(nr: INTEGER): STRING;
(* Programmdauer Stop minus Start aber Diff_TimeStamp24 *)

(* Falls m_oeglich, sollte Nr 0 nicht benutzt werden, sondern 1..10
   (Kompatibilit_aet zu anderen Units, wo Funktionen ohne Nr auf die 0 zurckgefhrt werden)
   Hier ist die 0 sogar anderweitig belegt :-) *)

PROCEDURE Steche_Moment;
(* GetDate und GetTime *)

PROCEDURE Konvert_Timestamp(t: STRING);
(* erstellt den Moment aus String timestamp *)

PROCEDURE Parse_Timestamp(t: STRING);
(* erstellt den Moment aus String timestamp mittels Watch.parse *)

PROCEDURE Formate_Moment(M: Moment);
(* Formatiert die Werte im Moment *)

PROCEDURE Zeige_Formate;
(* Zeigt die formatierten Werte im Moment an *)

PROCEDURE Lese_W_Moment(VAR M: Moment);
(* weist die "normalen" Variablen den Word zu *)

PROCEDURE Schreibe_W_Moment(VAR M: Moment);
(* schreibt die Word in die "normalen" Variablen *)

{$IFNDEF BIT_64}
PROCEDURE Normiere_Moment(VAR M: Moment);
(* Normieren des Momentes nach Ver_aenderung *)

PROCEDURE Dezimale_Moment(VAR M: Moment);
(* Umrechnen Momente in absolute Zahlen *)

FUNCTION Gueltig_Moment(VAR M: Moment): BOOLEAN;
(* Stehen im Moment gueltige Werte? *)
{$ENDIF}

PROCEDURE Entpacke_Moment(VAR M: Moment);
(* rechnet die absoluten MS um in einen Zeitpunkt *)

FUNCTION Gueltig_Werte(gjahr, gmonat, gtag, gstunde,
                       gminute, gsekunde, gmsek: LONGINT): boolean;
(* Sind diese Werte gueltig? *)

FUNCTION  Historisch_Test_Moment: BOOLEAN;
(* Ist Test_Moment < Akt_Moment *)

PROCEDURE Manuell_Moment(VAR M: Moment; mjahr, mmonat, mtag,
                         mstunde, mminute, msekunde, mmsek: LONGINT);
(* dem Moment die Werte zuweisen *)

PROCEDURE Addiere_Moment(VAR M: Moment;
                         ajahr, amonat, atag, astunde, aminute, asekunde, amsek: LONGINT);
(* dem Moment die Werte aufaddieren *)

PROCEDURE Differenz_Moment;
(* Diff_Moment = Akt_Moment - Test_Moment *)

{$IFNDEF BIT_64}
PROCEDURE SetzeNull_Summe_Moment;
(* Setzt Summe_Moment auf Null *)

PROCEDURE Addiere_Summe_Moment;
(* Summe_Moment = Summe_Moment + Diff_Moment *)
{$ENDIF}

IMPLEMENTATION

{$IFDEF DOS} USES TickTack; {$ENDIF}

CONST
  Wochen_Tage: ARRAY [0..7] OF STRING[10] =
    ('Sonntag', 'Montag', 'Dienstag', 'Mittwoch', 'Donnerstag',
     'Freitag', 'Sonnabend', 'Sonntag');

  Monatsnamen: ARRAY [1..12] OF STRING[9] =
    ('Januar', 'Februar', 'Maerz', 'April','Mai','Juni','Juli',
     'August','September', 'Oktober', 'November','Dezember');

{$IFDEF BIT_64} EinTag: CARD = 1000 * 60 * 60 * 24; {$ENDIF}

VAR
  Akt_Momente, Test_Momente: ARRAY[0..10] OF Moment;

FUNCTION schalt(jahr: LONGINT): INTEGER;
BEGIN
  schalt:=Ord(IstSchaltjahr(jahr));
END (* schalt *);

(* weist die "normalen" Variablen den Word zu *)
PROCEDURE Lese_W_Moment(VAR M: Moment);
BEGIN
  WITH M DO BEGIN
    wjahr:=  jahr;   wmonat:= monat;  wtag:=    tag;     wwotag:=wotag;
    wstunde:=stunde; wminute:=minute; wsekunde:=sekunde; wmsek:= msek;
  END;
END;

(* schreibt die Word in die "normalen" Variablen *)
PROCEDURE Schreibe_W_Moment(VAR M: Moment);
BEGIN
  WITH M DO BEGIN
    jahr:=  wjahr;   monat:= wmonat;  tag:=    wtag;     wotag:=wwotag;
    stunde:=wstunde; minute:=wminute; sekunde:=wsekunde; msek:= wmsek;
  END;
END;

(* erzeuge die TS f_uer Dateizugriff Dos und Unix *)
PROCEDURE DosUnx_TS(VAR M: Moment);
BEGIN
  WITH M DO BEGIN
    {$IFDEF BIT_64}
    DecodeDateTime(AbsDT, wjahr, wmonat, wtag, wstunde, wminute, wsekunde, wmsek);
    Schreibe_W_Moment(M);
    {$ENDIF}
    Lese_W_Moment(M);
    {$IFNDEF BIT_64} wmsek := 10 * wmsek; {$ENDIF}
    TSDos:=DTWerteZuTS(wjahr, wmonat, wtag, wstunde, wminute, wsekunde);
    TSUnx:=DTWerteZuTSUnx(wjahr, wmonat, wtag, wstunde, wminute, wsekunde, wmsek);
    TSWin:=TSUnxZuTSWin(TSUnx);
  END;
END (* DosUnx *);

{$IFNDEF BIT_64}
PROCEDURE Normiere_Moment(VAR M: Moment);
VAR
  {tjahr,} tmonat, {ttag,} tstunde, tminute, tsekunde, tmsek: LONGINT;
BEGIN
  WITH M DO BEGIN
    IF jahr < 1 THEN  EXIT;
    (* wenn wir hier negative Jahre angeboten bekommen,
       sind wir mit der Wissenschaft am Ende *)
    WHILE msek < 0 DO BEGIN
      msek:=msek + 100; DEC(sekunde);
    END;
    WHILE sekunde < 0 DO BEGIN
      sekunde:=sekunde + 60; DEC(minute);
    END;
    WHILE minute < 0 DO BEGIN
      minute:=minute+60; DEC(stunde);
    END;
    WHILE stunde < 0 DO BEGIN
      stunde:=stunde + 24; DEC(tag);
    END;
    WHILE monat < 1 DO BEGIN
      monat:=monat + 12; DEC(jahr);
      IF jahr<1 THEN  EXIT;
    END;

    WHILE tag < 1 DO BEGIN
      tag:=tag + TageImMonat(jahr, monat-1);
      DEC(monat);
      IF monat < 1 THEN  BEGIN
        monat:=12; DEC(jahr);
        IF jahr < 1 THEN  EXIT;
      END;
    END;

    tmsek:=    msek;
    msek:=     msek    MOD 100;
    sekunde:=  sekunde + tmsek div 100;
    tsekunde:= sekunde;
    sekunde:=  sekunde MOD 60;
    minute:=   minute  + tsekunde div 60;
    tminute:=  minute;
    minute:=   minute  MOD 60;
    stunde:=   stunde  + tminute  div 60;
    tstunde:=  stunde;
    stunde:=   stunde  MOD 24;
    tag:=      tag     + tstunde  div 24;
    tmonat:=   monat;
    monat:=    monat   MOD 12;  IF monat = 0 THEN  monat:= 12;
    jahr:=     jahr    + ((tmonat - monat) div 12);

    WHILE (tag > TageImMonat(jahr, monat)) DO BEGIN
      tag:=tag-TageImMonat(jahr, monat);
      IF monat < 12
        THEN  INC(monat)
        ELSE BEGIN monat:=1; INC(jahr); END;
    END;
    Dezimale_Moment(M);
    wotag:=(abstag MOD 7) + 1;
  END;
END  (* Normiere_Moment *);

PROCEDURE Dezimale_Moment(VAR M: Moment);
VAR
  dmonate,djahre: LONGINT;
BEGIN
  WITH M DO BEGIN
    abstag:= tag - 1;
    dmonate:=monat - 1;
    djahre:= jahr - 1;

    WHILE dmonate > 0 DO BEGIN
      abstag:=abstag + TageImMonat(jahr, dmonate);
      DEC(dmonate);
    END;
    WHILE djahre > 0 DO BEGIN
      abstag:=abstag + 365 + schalt(djahre);
      DEC(djahre);
    END;

    absMS:=abstag;
    absMS:=absMS   * 8640000 +
           stunde  *  360000 +
           minute  *    6000 +
           sekunde *     100 +
           msek;
    abssek:=INT(absMS/100.00);
    wotag:=(abstag MOD 7) + 1; 
    DosUnx_TS(Akt_Moment);    
  END;
END (* Dezimale_Moment *);

(* rechnet die absoluten MS um in einen Zeitpunkt *)
PROCEDURE Entpacke_Moment(VAR M: Moment);
VAR
  etag, nbis : LONGINT;
  sicher, {$IFDEF ZEITOPT} rest {$ELSE} schwelle {$ENDIF} : DOUBLE;
BEGIN
  WITH M DO BEGIN
    sicher:=absMS;
  {$IFDEF ZEITOPT}
    (* (hoffentlich) verlustarmer Gleitkomma Modulus :-)
       ca 10% schneller als die do-whiles mit dem Schwellwert.
       Leider werden die TRUNC/ROUND-Zwischenergebnisse fr 16bit zu gro_sz,
       so da_sz dieses extra fr DOS entworfene Verfahren doch nicht verwendet
       werden kann unter Turbo-Pascal. *)
    rest:=(absMS / 100.0) - TRUNC(absMS / 100.0);
    msek:=ROUND(100.0 * rest);
    absMS:=(absMS - msek) / 100.0;
    rest:=(absMS / 60.0) - TRUNC(absMS / 60.0);
    sekunde:=ROUND(60.0 * rest);
    absMS:=(absMS - sekunde) / 60.0;
    rest:=(absMS / 60.0) - TRUNC(absMS / 60.0);
    minute:=ROUND(60.0 * rest);
    absMS:=(absMS - minute) / 60.0;
    rest:=(absMS / 24.0) - TRUNC(absMS / 24.0);
    stunde:=ROUND(24.0 * rest);
    absMS:=(absMS - stunde) / 24.0;
    etag:=TRUNC(absMS);
  {$ELSE}
    etag:=0;
    schwelle:=1000. * 24 * 3600 * 100.;
    WHILE absMS >= schwelle  DO BEGIN
      INC(etag, 1000); absMS:=absMS - schwelle;
    END;
    schwelle:=24 * 3600 * 100.;
    WHILE absMS >= schwelle  DO BEGIN
      INC(etag); absMS:=absMS - schwelle;
    END;
    stunde:=0;
    schwelle:=3600 * 100;
    WHILE absMS >= schwelle  DO BEGIN
      INC(stunde); absMS:=absMS - schwelle;
    END;
    minute:=0;
    schwelle:=60 * 100;
    WHILE absMS >= schwelle  DO BEGIN
      INC(minute); absMS:=absMS - schwelle;
    END;
    sekunde:=0;
    schwelle:=100;
    WHILE absMS >= schwelle  DO BEGIN
      INC(sekunde); absMS:=absMS - schwelle;  
    END;
    msek:=TRUNC(absMS);
  {$ENDIF}
    absMS:=sicher;
    jahr:=1; monat:=1; tag:=1;
    WHILE etag > 0 DO BEGIN
      nbis:=TageImMonat(jahr,monat) - tag + 1;
      IF nbis > etag THEN  nbis:=1;
      INC(tag, nbis); (* bevorzugt ganze Monate :-) *)
      IF tag > TageImMonat(jahr, monat) THEN  BEGIN
        INC(monat); tag:=1;
        IF monat > 12 THEN  BEGIN
          INC(jahr); monat:=1;
        END;
      END;
      DEC(etag, nbis);
    END;
  END;
  Dezimale_Moment(M);
END; (* Entpacke_Moment *)

{$ELSE}

(* entpacke das DT in Word und weise sie in die "normalen" Variablen *)
PROCEDURE Entpacke_Moment(VAR M: Moment);
BEGIN
  WITH M DO BEGIN
    DecodeDate(absDT, wjahr, wmonat, wtag);
    DecodeTime(absDT, wstunde, wminute, wsekunde, wmsek);

    wwotag:=DayOfWeek(absDT) - 1; IF wwotag=0 THEN  wwotag:=7;
    jahr:=  wjahr;   monat:= wmonat;  tag:=    wtag;     wotag:=wwotag;
    stunde:=wstunde; minute:=wminute; sekunde:=wsekunde; msek:= wmsek;
  END;
END;
{$ENDIF}

PROCEDURE Steche_Moment;
BEGIN
  WITH Akt_Moment DO BEGIN
  {$IFNDEF BIT_64}
    GetDate(wjahr, wmonat, wtag, wwotag);
    {$IFDEF DOS}
      GetTimeDOS(wstunde, wminute, wsekunde, wmsek);
    {$ELSE}
      GetTime(wstunde, wminute, wsekunde, wmsek);
    {$ENDIF}
      Schreibe_W_Moment(Akt_Moment);
      Dezimale_Moment(Akt_Moment);
  {$ELSE}
    absDT:=now;
    absTS:=DateTimeToTimeStamp(absDT);
    absMS:=CARD(TimeStampToMSecs(absTS));
  {$ENDIF}
  DosUnx_TS(Akt_Moment);
  END;
END (* Steche_Moment *);

FUNCTION StopUhr(Kommentar: STRING): STRING;
VAR
  KM: STRING;
BEGIN
  KM:=Kommentar; IF KM=':' THEN  KM:='StopUhr:';
  Steche_Moment;
  Formate_Moment(Akt_Moment);
  StopUhr:=KM+Formate.TimeStamp;
END (* Stop_Uhr *);

FUNCTION Intervall_Start(nr: INTEGER): STRING;
BEGIN
  Intervall_Start:=StopUhr('Start:');
  Akt_Momente[nr]:=Akt_Moment;
  Test_Momente[nr]:=Akt_Moment;
END;

FUNCTION Intervall_Stop(nr: INTEGER): STRING;
BEGIN
  Intervall_Stop:=StopUhr('Stop: ');
  Akt_Momente[nr]:=Akt_Moment;
  Test_Moment:=Test_Momente[nr];
END;

FUNCTION Intervall_Dauer(nr: INTEGER):STRING;
VAR dauer: STRING;
BEGIN
  Akt_Momente[0]:=Akt_Moment; Test_Momente[0]:=Test_Moment;
  Akt_Moment:=Akt_Momente[nr]; Test_Moment:=Test_Momente[nr];
  Differenz_Moment;
  IF Diff_Moment.rsekunde >= 0
    THEN  dauer:=COPY(Diff_TimeStamp, 10, 255)
    ELSE dauer:='Negativ??? :-)';
  IF Diff_Moment.rtag >= 1
    THEN  dauer:=StrLong(Trunc(Diff_Moment.rtag))+'T '+dauer;
  Intervall_dauer:='Dauer:'+dauer;
  Akt_Moment:=Akt_Momente[0]; Test_Moment:=Test_Momente[0];
END;

FUNCTION Intervall_Dauer24(nr: INTEGER):STRING;
BEGIN
  Akt_Momente[0]:=Akt_Moment; Test_Momente[0]:=Test_Moment;
  Akt_Moment:=Akt_Momente[nr]; Test_Moment:=Test_Momente[nr];
  Differenz_Moment;
  Intervall_dauer24:=Diff_TimeStamp24;
  Akt_Moment:=Akt_Momente[0]; Test_Moment:=Test_Momente[0];
END;

PROCEDURE Konvert_Timestamp(t: STRING);
VAR
  {$IFNDEF BIT_64}
  tjahr, tmonat, ttag, tstunde, tminute, tsekunde, tmsek: LONGINT;
  w: WORD;
  {$ELSE}
  i, k, e: WORD;
  {$ENDIF}
BEGIN
  {$IFNDEF BIT_64}
  VAL(copy(t,1,2),  ttag, w);     VAL(copy(t,4,2),  tmonat, w);  VAL(copy(t,7,4),  tjahr, w);
  VAL(copy(t,12,2), tstunde, w);  VAL(copy(t,15,2), tminute, w); VAL(copy(t,18,2), tsekunde ,w);
  VAL(copy(t,21,2), tmsek, w);
  IF w=0 THEN  Manuell_Moment(Akt_Moment, tjahr, tmonat, ttag, tstunde, tminute, tsekunde, tmsek);
  {$ELSE}
  WITH Akt_Moment DO BEGIN
    (* StrToDateTime geht schlampig mit den ms um,
       daher: machen wir sie hier mindestens dreistellig *)
    k:=POS(',', t); e:=length(t);
    IF k > 0 THEN
      FOR i:=e+1 to k+3 DO t:=t+'0';
    absDT:=StrToDateTime(t);
    absTS:=DateTimeToTimeStamp(absDT);
    absMS:=CARD(TimeStampToMSecs(absTS));
  END;
  {$ENDIF}
  DosUnx_TS(Akt_Moment);
  Formate_Moment(Akt_Moment);
END;

PROCEDURE Parse_Timestamp(t: STRING);
VAR
  DT: TDosTime;
  UT: TUnxTime;
  tjahr, tmonat, ttag, tstunde, tminute, tsekunde, tmsek: WORD;
BEGIN
  DT:=0; UT:=0;
  ParseDT(t, DT, UT);
  IF UT > 0  THEN BEGIN
    tjahr:=0; tmonat:=0; ttag:=0; tstunde:=0; tminute:=0; tsekunde:=0; tmsek:=0;
    TSUnxZuDTWerte(UT, tjahr, tmonat, ttag, tstunde, tminute, tsekunde, tmsek);
    Manuell_Moment(Akt_Moment, tjahr, tmonat, ttag, tstunde, tminute, tsekunde, tmsek);
    Formate_Moment(Akt_Moment);
    Write(t, ' ergibt: ')
  END
  ELSE WITH Akt_Moment DO BEGIN
    Write(t, ' ist leider ung'+_ue+'ltig: ', StopUhr(':'), ' ');
    DT:=TSDos;
    UT:=TSUnx;
  END;
  Writeln('(DOS) ', strCard(DT), ' = ', FormatTS(DT), ' (UNIX) ', strCard(UT), ' = ', FormatTSUnx(UT));
END;

PROCEDURE Formate_Moment(M: Moment);
{$IFDEF BIT_64}
VAR tjahr, tmonat, ttag, tstunde, tminute, tsekunde, tmsek: WORD;
{$ENDIF}
BEGIN
  WITH M DO BEGIN
    {$IFNDEF BIT_64} Lese_W_Moment(M);
    {$ELSE}          Entpacke_Moment(M);
    {$ENDIF}
    Formate.Wochentag    := Wochen_Tage[wwotag];
    Formate.TT_MM_JJ     := str0(wtag, 2) +'.'+
                            str0(wmonat, 2) +'.'+
                            str0(wjahr MOD 100, 2);
    Formate.TT_MM_JJJJ   := str0(wtag ,2) + '.'+
                            str0(wmonat, 2) +'.'+
                            str0(wjahr, 4);
    Formate.JJJJ_MM_TT   := str0(wjahr, 4) +'-'+
                            str0(wmonat, 2) +'-'+
                            str0(wtag, 2);
    Formate.HH_MM_SS     := str0(wstunde, 2) +':'+
                            str0(wminute, 2) +':'+
                            str0(wsekunde, 2);
    Formate.HH_MM_SS_MS  := str0(wstunde, 2) +':'+
                            str0(wminute, 2) +':'+
                            str0(wsekunde, 2) +','+
                            str0(wmsek, {$IFNDEF BIT_64} 2 {$ELSE} 3 {$ENDIF} );
    Formate.Timestamp    := str0(wtag, 2) +'.'+
                            str0(wmonat, 2) +'.'+
                            str0(wjahr, 4) +' '+
                            Formate.HH_MM_SS_MS;
    Formate.FileStampDT  := 'd'+str0(M.jahr,4)+
                            str0(M.monat,2)+
                            str0(M.tag,2)+
                            '_t'+str0(M.stunde,2)+
                            str0(M.minute,2)+
                            str0(M.sekunde,2)+
                            str0(M.msek,2);
    Formate.DdatumTzeit  := 'D'+str0(wjahr, 4)+
                            str0(wmonat, 2)+
                            str0(wtag, 2)+
                            'T'+str0(wstunde, 2)+
                            str0(wminute, 2)+
                            str0(wsekunde, 2)+
                            str0(wmsek, {$IFNDEF BIT_64} 2 {$ELSE} 3 {$ENDIF} );
    Formate.Ddatum       := 'D'+str0(wjahr, 4)+
                            str0(wmonat, 2)+
                            str0(wtag, 2);
    Formate.Datum_Zeit   := Formate.Wochentag +', '+
                            str0(wtag, 2) +'.'+
                            Monatsnamen[wmonat] +' '+
                            str0(wjahr, 4) +', ' +
                            Formate.HH_MM_SS;
    Formate.Abs_Tag      := {$IFNDEF BIT_64} StrLong(abstag);
                            {$ELSE}          StrLong(absTS.date-1);
                            {$ENDIF}
    Formate.Abs_MS       := {$IFDEF DOS}         strdouble(absMS);
                            {$ELSE}
                                {$IFNDEF BIT_64} strsoftreal(absMS);
                                {$ELSE}          strCard(absMS-EinTag);
                                {$ENDIF}
                            {$ENDIF}
    {$IFDEF BIT_64}
    DecodeDate(AbsDT, tjahr, tmonat, ttag);
    DecodeTime(AbsDT, tstunde, tminute, tsekunde, tmsek);
    Formate.Abs_DT       := StrDouble(AbsDT)+' = '+str0(ttag, 2)+'.'+str0(tmonat, 2)+'.'+str0(tjahr, 4)+
                            ' '+str0(tstunde, 2)+':'+str0(tminute, 2)+':'+str0(tsekunde, 2)+','+str0(tmsek, 3);
    DecodeDate(TimeStampToDateTime(AbsTS), tjahr, tmonat, ttag);
    DecodeTime(TimeStampToDateTime(AbsTS), tstunde, tminute, tsekunde, tmsek);
    Formate.Abs_TS       := strCard(AbsTS.Date)+' / '+strCard(AbsTS.Time)+' = '+str0(ttag, 2)+'.'+str0(tmonat, 2)+'.'+
                            str0(tjahr, 4) +' '+str0(tstunde, 2)+':'+str0(tminute, 2)+':'+str0(tsekunde, 2)+','+str0(tmsek, 3);
    {$ENDIF}
    Formate.TS_Dos       := StrCard(TSDos)+' = '+FormatTS(TSDos);
    Formate.TS_Unx       := StrCard(TSUnx)+' = '+FormatTSUnx(TSUnx);
    Formate.TS_Win       := StrDouble(TSWin)+' = '+FormatTSWin(TSWin);
  END;
END (* Formate_Moment *);

PROCEDURE Zeige_Formate;
BEGIN
  WriteLn('Wochentag     = ',Formate.Wochentag);
  WriteLn('TT.MM.JJ      = ',Formate.TT_MM_JJ);
  WriteLn('TT.MM.JJJJ    = ',Formate.TT_MM_JJJJ);
  WriteLn('JJJJ-MM-TT    = ',Formate.JJJJ_MM_TT);
  WriteLn('HH:MM:SS      = ',Formate.HH_MM_SS);
  WriteLn('HH:MM:SS,MS   = ',Formate.HH_MM_SS_MS);
  WriteLn('Timestamp     = ',Formate.Timestamp);
  WriteLn('Datum/Zeit    = ',Formate.Datum_Zeit);
  WriteLn('absolute Zeit = ',Formate.Abs_Tag,' Tage == ',
                             Formate.Abs_ms, {$IFNDEF BIT_64} ' HS' {$ELSE} ' ms' {$ENDIF} );
  {$IFDEF BIT_64}
  WriteLn('TS_Abs_DT     = ',Formate.Abs_DT);
  WriteLn('TS_Abs_TS     = ',Formate.Abs_TS);
  {$ENDIF};
  WriteLn('TS_Dos        = ',Formate.TS_Dos);
  WriteLn('TS_Unx        = ',Formate.TS_Unx);
  WriteLn('TS_Win        = ',Formate.TS_Win);
END (* Zeige_Format *);

{$IFNDEF BIT_64}
(* Diese Funktion ist exclusiv fr TOLLZEIT :-) *)
FUNCTION Gueltig_Moment(VAR M: Moment): BOOLEAN;
BEGIN
  WITH M DO Gueltig_Moment:= Gueltig_Werte(jahr, monat, tag, stunde, minute, sekunde, msek);
END (* Gueltig_Moment *);
{$ENDIF}

FUNCTION Gueltig_Werte(gjahr, gmonat, gtag, gstunde, gminute, gsekunde, gmsek: LONGINT): BOOLEAN;
BEGIN
  Gueltig_Werte:=true;
  IF gjahr <= 0 THEN  Gueltig_Werte:=false;
  IF (gtag < 1) or
     (gmonat < 1) OR (gmonat > 12) OR
     (gtag > TageImMonat(gjahr, gmonat)) THEN  Gueltig_Werte:=false;
  IF (gstunde  < 0) OR (gstunde  > 23)  THEN  Gueltig_Werte:=false;
  IF (gminute  < 0) OR (gminute  > 59)  THEN  Gueltig_Werte:=false;
  IF (gsekunde < 0) OR (gsekunde > 59)  THEN  Gueltig_Werte:=false;
  IF (gmsek    < 0) OR (gmsek    > 99)  THEN  Gueltig_Werte:=false;
END (* Gueltig_Werte *);

FUNCTION Historisch_Test_Moment: BOOLEAN;
BEGIN
  {$IFNDEF BIT_64}
  Dezimale_Moment(Test_Moment);
  Dezimale_Moment(Akt_Moment);
  {$ENDIF}
  Historisch_Test_Moment:=(Test_Moment.absMS < Akt_Moment.absMS);
END (* Historisch_Moment *);

PROCEDURE Manuell_Moment(VAR M: Moment;
                         mjahr, mmonat, mtag, mstunde, mminute, msekunde, mmsek: LONGINT);
BEGIN
  WITH M DO BEGIN
    {$IFNDEF BIT_64}
    jahr:=  mjahr;   monat:= mmonat;  tag:=    mtag;
    stunde:=mstunde; minute:=mminute; sekunde:=msekunde; msek:=mmsek;
    Normiere_Moment(M);
    {$ELSE}
    absDT:=StrToDateTime(Str0Long(mtag, 2) +'.'+
           Str0Long(mmonat, 2) +'.'+
           Str0Long(mjahr, 4) +' '+
           Str0Long(mstunde, 2) +':'+
           Str0Long(mminute, 2) +':'+
           Str0Long(msekunde, 2) +','+
           Str0Long(mmsek, 3));
    absTS:=DateTimeToTimeStamp(absDT);
    absMS:=CARD(TimeStampToMSecs(absTS));
    {$ENDIF}
    DosUnx_TS(Akt_Moment);
  END;
END (* Manuell_Moment *);

PROCEDURE Addiere_Moment(VAR M: Moment;
                         ajahr, amonat, atag, astunde, aminute, asekunde, amsek: LONGINT);
BEGIN
  WITH M DO BEGIN
    {$IFNDEF BIT_64}
    IF ajahr <> 0    THEN  jahr:=   jahr    + ajahr;
    IF amonat <> 0   THEN  monat:=  monat   + amonat;
    IF atag <> 0     THEN  tag:=    tag     + atag;
    IF astunde <> 0  THEN  stunde:= stunde  + astunde;
    IF aminute <> 0  THEN  minute:= minute  + aminute;
    IF asekunde <> 0 THEN  sekunde:=sekunde + asekunde;
    IF amsek <> 0    THEN  msek:=   msek    + amsek;
    Normiere_Moment(M);
    {$ELSE}
    IF ajahr <> 0    THEN  absDT:=IncYear(absDT, ajahr);
    IF amonat <> 0   THEN  absDT:=IncMonth(absDT, amonat);
    IF atag <> 0     THEN  absDT:=IncDay(absDT, atag);
    IF astunde <> 0  THEN  absDT:=IncHour(absDT, astunde);
    IF aminute <> 0  THEN  absDT:=IncMinute(absDT, aminute);
    IF asekunde <> 0 THEN  absDT:=IncSecond(absDT, asekunde);
    IF amsek <> 0    THEN  absDT:=IncMilliSecond(absDT, amsek);
    absTS:=DateTimeToTimeStamp(absDT);
    absMS:=CARD(TimeStampToMSecs(absTS));
    {$ENDIF}
    DosUnx_TS(Akt_Moment);
  END;
END (* Addiere_Moment *);

PROCEDURE Differenz_Moment;
VAR reverse: Boolean;
    von, bis: Moment;
BEGIN
  {$IFDEF BIT_64}
  Entpacke_Moment(Akt_Moment);
  Entpacke_Moment(Test_Moment);
  {$ENDIF}
  IF Historisch_Test_Moment
    THEN  BEGIN
      reverse:=false;
      von:=Test_Moment; bis:=Akt_Moment;
    END
    ELSE BEGIN
      reverse:=true;
      von:=Akt_Moment; bis:=Test_Moment;
    END;

  WITH Diff_Moment DO BEGIN
    IF bis.msek > von.msek-1
      THEN  imsek:=bis.msek - von.msek
      ELSE BEGIN
        imsek:=bis.msek - von.msek + {$IFNDEF BIT_64} 100 {$ELSE} 1000 {$ENDIF} ;
        DEC(bis.sekunde)
      END;
    IF bis.sekunde > von.sekunde-1
      THEN  isekunde:=bis.sekunde - von.sekunde
      ELSE BEGIN
        isekunde:=bis.sekunde + 60 - von.sekunde;
        DEC(bis.minute)
      END;
    IF bis.minute > von.minute-1
      THEN  iminute:=bis.minute - von.minute
      ELSE BEGIN
        iminute:=bis.minute + 60 - von.minute;
        DEC(bis.stunde)
      END;
    IF bis.stunde > von.stunde-1
      THEN  istunde:=bis.stunde - von.stunde
      ELSE BEGIN
        istunde:=bis.stunde + 24 - von.stunde;
            DEC(bis.tag)
      END;

    IF bis.tag > von.tag-1
      THEN  itag:=bis.tag - von.tag
      ELSE BEGIN
        IF reverse
          THEN itag:=bis.tag + TageImMonat(von.jahr, von.monat) - von.tag
          ELSE itag:=bis.tag + TageImMonat(bis.jahr, bis.monat-1) - von.tag;
        DEC(bis.monat)
      END;
    IF bis.monat > von.monat-1
      THEN  imonat:=bis.monat - von.monat
      ELSE BEGIN
        imonat:=bis.monat+ 12 -von.monat;
        DEC(bis.jahr)
      END;

    ijahr:=bis.jahr - von.jahr;

    IF reverse THEN  BEGIN
      ijahr:=-ijahr; imonat:=-imonat; itag:=-itag;
      istunde:=-istunde; iminute:=-iminute; isekunde:=-isekunde; imsek:=-imsek;
    END;
    rsekunde:= Akt_Moment.absMS - Test_Moment.absMS;
    rsekunde:= rsekunde / {$IFNDEF BIT_64} 100.00 {$ELSE} 1000.0 {$ENDIF} ;
    rminute:=  rsekunde / 60.00;
    rstunde:=  rminute  / 60.00;
    rtag:=     rstunde  / 24.00;
    rjahr:=    rtag     / 365.22;
    rmonat:=   rjahr    * 12.00;
    rwoche:=   rtag     / 7.00;

    IF rsekunde >= 0
      THEN BEGIN
        Diff_TimeStamp:=  Str0Long(itag, 2) +'.'+
                          Str0Long(imonat, 2) +'.'+
                          Str0Long(ijahr, 2) +' '+
                          Str0Long(istunde, 2) +':'+
                          Str0Long(iminute, 2) +':'+
                          Str0Long(isekunde, 2) +','+
                          Str0Long(imsek, {$IFNDEF BIT_64} 2 {$ELSE} 3 {$ENDIF} );
        (* die Monate/Jahre vernachl_aessigen wir hier bis zum Beweis des Gegenteils *)
        Diff_TimeStamp24:=Str0Long(istunde + 24*itag, 2) +':'+
                          Str0Long(iminute, 2) +':'+
                          Str0Long(isekunde, 2) +','+
                          Str0Long(imsek, {$IFNDEF BIT_64} 2 {$ELSE} 3 {$ENDIF} );
        Diff_Einzel:=     Str0Long(ijahr, 2) +' Jahre '+
                          Str0Long(imonat ,2) +' Monate '+
                          Str0Long(itag, 2) +' Tage, '+
                          Str0Long(istunde, 2) +':'+
                          Str0Long(iminute, 2) +':'+
                          Str0Long(isekunde, 2) +','+
                          Str0Long(imsek, {$IFNDEF BIT_64} 2 {$ELSE} 3 {$ENDIF} );
      END
      ELSE BEGIN
        Diff_TimeStamp:=  StrLong(itag) +'/'+
                          StrLong(imonat) +'/'+
                          StrLong(ijahr) +' '+
                          StrLong(istunde) +'/'+
                          StrLong(iminute) +'/'+
                          StrLong(isekunde )+'/'+
                          StrLong(imsek);
        (* die Monate/Jahre vernachl_aessigen wir hier bis zum Beweis des Gegenteils *)
        Diff_TimeStamp24:=StrLong(istunde + 24*itag) +'/'+
                          StrLong(iminute) +'/'+
                          StrLong(isekunde) +'/'+
                          StrLong(imsek);
        Diff_Einzel:=     StrLong(ijahr) +' Jahre '+
                          StrLong(imonat) +' Monate '+
                          StrLong(itag) +' Tage, '+
                          StrLong(istunde) +'/'+
                          StrLong(iminute) +'/'+
                          StrLong(isekunde) +'/'+
                          StrLong(imsek);
      END;

    {$IFDEF DOS}
    Diff_Abs_Sekunden:=strkdouble(rsekunde,2) +' s';
    Diff_Abs_Jahre:=strkdouble(rjahr,3) +' Jahre';
    Diff_Abs_Monate:=strkdouble(rmonat,3) +' Monate';
    Diff_Abs_Wochen:=strkdouble(rwoche,3) +' Wochen';
    Diff_Abs_Tage:=strkdouble(rtag,3)+' Tage';
    Diff_Abs_Stunden:=strkdouble(rstunde,3) +' h';
    Diff_Abs_Minuten:=strkdouble(rminute,3) +' min';
    {$ELSE}
    Diff_Abs_Sekunden:=StrkSoftReal(rsekunde, {$IFNDEF BIT_64} 2 {$ELSE} 3 {$ENDIF} ) +' s';
    Diff_Abs_Jahre:=StrkSoftReal(rjahr,3) +' Jahre';
    Diff_Abs_Monate:=StrkSoftReal(rmonat,3) +' Monate';
    Diff_Abs_Wochen:=StrkSoftReal(rwoche,3) +' Wochen';
    Diff_Abs_Tage:=StrkSoftReal(rtag,3)+' Tage';
    Diff_Abs_Stunden:=StrkSoftReal(rstunde,3) +' h';
    Diff_Abs_Minuten:=StrkSoftReal(rminute,3) +' min';
    {$ENDIF}
    Diff_Abs:=Diff_Abs_Jahre +' / '+
              Diff_Abs_Monate +' / '+
              Diff_Abs_Wochen +' / '+
              Diff_Abs_Tage +' / '+
              Diff_Abs_Stunden +' / '+
              Diff_Abs_Minuten +' / '+
              Diff_Abs_Sekunden;
  END; (* with Diff_Moment *)
END (* Differenz_Moment *);

{$IFNDEF BIT_64}
PROCEDURE Addiere_Summe_Moment;
BEGIN
  WITH Summe_Moment DO BEGIN
    imsek   := imsek    + Diff_Moment.imsek;
    isekunde:= isekunde + Diff_Moment.isekunde;
    iminute := iminute  + Diff_Moment.iminute;
    istunde := istunde  + Diff_Moment.istunde;
    itag    := itag     + Diff_Moment.itag;
    imonat  := imonat   + Diff_Moment.imonat;
    ijahr   := ijahr    + Diff_Moment.ijahr;
    WHILE imsek>99 DO BEGIN
      DEC(imsek,100); INC(isekunde);
    END;
    WHILE isekunde>59 DO BEGIN
      DEC(isekunde,60); INC(iminute);
    END;
    WHILE iminute>59 DO BEGIN
      DEC(iminute,60); INC(istunde);
    END;
    WHILE istunde>23 DO BEGIN
      DEC(istunde,24); INC(itag);
    END;
    WHILE itag>30 DO BEGIN
      DEC(itag,31); INC(imonat);
    END;
    WHILE imonat>11 DO BEGIN
      DEC(imonat,12); INC(ijahr);
    END;
    rsekunde:= imsek    * 1.0;
    rsekunde:= rsekunde / 100.00;
    rminute:=  rsekunde / 60.00;
    rstunde:=  rminute  / 60.00;
    rtag:=     rstunde  / 24.00;
    rjahr:=    rtag     / 365.22;
    rmonat:=   rjahr    * 12.00;
    rwoche:=   rtag     / 7.00;
    Summe_TimeStamp:=  Str0Long(itag, 2) +'.'+
                       Str0Long(imonat, 2) +'.'+
                       Str0Long(ijahr, 4) +' '+
                       Str0Long(istunde, 2) +':'+
                       Str0Long(iminute, 2) +':'+
                       Str0Long(isekunde, 2) +','+
                       Str0Long(imsek,2);
    (* die Monate/Jahre vernachl_aessigen wir hier bis zum Beweis des Gegenteils *)
    Summe_TimeStamp24:=Str0Long(istunde+ 24*itag, 2) +':'+
                       Str0Long(iminute, 2) +':'+
                       Str0Long(isekunde, 2) +','+
                       Str0Long(imsek, 2);
   END;
END (* Addiere_Summe_Moment *);

PROCEDURE SetzeNull_Summe_Moment;
BEGIN
  WITH Summe_Moment DO BEGIN
    imsek:= 0; isekunde:= 0; iminute:= 0; istunde:= 0;
    itag:= 0;  imonat:= 0;   ijahr:= 0;
    rsekunde:= 0.0; rminute:= 0.0; rstunde:= 0.0;
    rtag:= 0.0;     rjahr:= 0.0;   rmonat:= 0.0; rwoche:= 0.0;
    Summe_TimeStamp:='00.00.0000 00:00:00,00';
    (* die Monate/Jahre vernachl_aessigen wir hier bis zum Beweis des Gegenteils *)
    Summe_TimeStamp24:='00:00:00,00';
  END;
END (* SetzeNull_Summe_Moment *);
{$ENDIF}

BEGIN
  {$IFDEF DOS} Setkorrektur(-130);
  {$ENDIF}
  {$IFDEF BIT_64}    (* Writeln('Unit Zeit 64bit-Modulus aktiv');            *) 
  {$ELSE}            (* Writeln('Unit Zeit 64bit-Modulus nicht aktiv');      *)
  {$ENDIF}
  {$IFDEF ZEITOPT}   (* Writeln('Unit Zeit Gleitkomma-Modulus aktiv');       *)
  {$ELSE}            (* Writeln('Unit Zeit Gleitkomma-Modulus nicht aktiv'); *)
  {$ENDIF}
  Monatsnamen[3]:='M'+_ae+'rz';
END (* Zeit.TPU *).
